@@ -601,8 +601,7 @@ pango_attr_strikethrough_new (gboolean strikethrough)
 /**
  * pango_attr_rise_new:
  * @rise: the amount that the text should be displaced vertically,
- *        in 10'000ths of an em. Positive values displace the
- *        text upwards.
+ *        in Pango units. Positive values displace the text upwards.
  * 
  * Create a new baseline displacement attribute.
  * 
@@ -643,6 +642,57 @@ pango_attr_scale_new (double scale_factor)
   return pango_attr_float_new (&klass, scale_factor);
 }
 
+/**
+ * pango_attr_fallback_new:
+ * @enable_fallback: %TRUE if we should fall back on other fonts 
+ *                   for characters the active font is missing.
+ * 
+ * Create a new font fallback attribute.
+ *
+ * If fallback is disabled, characters will only be used from the
+ * closest matching font on the system. No fallback will be done to
+ * other fonts on the system that might contain the characters in the
+ * text.
+ * 
+ * Return value: the new #PangoAttribute.
+ **/
+PangoAttribute *
+pango_attr_fallback_new (gboolean enable_fallback)
+{
+  static const PangoAttrClass klass = {
+    PANGO_ATTR_FALLBACK,
+    pango_attr_int_copy,
+    pango_attr_int_destroy,
+    pango_attr_int_equal,
+  };
+
+  return pango_attr_int_new (&klass, (int)enable_fallback);
+}
+
+/**
+ * pango_attr_letter_spacing_new:
+ * @letter_spacing: amount of extra space to add between graphemes
+ *   of the text, in Pango units.
+ * 
+ * Create a new letter-spacing attribute.
+ * 
+ * Return value: the new #PangoAttribute.
+ *
+ * Since: 1.6
+ **/
+PangoAttribute *
+pango_attr_letter_spacing_new (int letter_spacing)
+{
+  static const PangoAttrClass klass = {
+    PANGO_ATTR_LETTER_SPACING,
+    pango_attr_int_copy,
+    pango_attr_int_destroy,
+    pango_attr_int_equal
+  };
+
+  return pango_attr_int_new (&klass, letter_spacing);
+}
+
 static PangoAttribute *
 pango_attr_shape_copy (const PangoAttribute *attr)
 {
@@ -941,8 +991,14 @@ pango_attr_list_change (PangoAttrList  *list,
   GSList *tmp_list, *prev, *link;
   gint start_index = attr->start_index;
   gint end_index = attr->end_index;
-  
+
   g_return_if_fail (list != NULL);
+
+  if (start_index == end_index)	/* empty, nothing to do */
+    {
+      pango_attribute_destroy (attr);
+      return;
+    }
   
   tmp_list = list->attributes;
   prev = NULL;
@@ -1274,8 +1330,11 @@ pango_attr_iterator_next (PangoAttrIterator *iterator)
   while (iterator->next_attribute &&
 	 ((PangoAttribute *)iterator->next_attribute->data)->start_index == iterator->start_index)
     {
-      iterator->attribute_stack = g_list_prepend (iterator->attribute_stack, iterator->next_attribute->data);
-      iterator->end_index = MIN (iterator->end_index, ((PangoAttribute *)iterator->next_attribute->data)->end_index);
+      if (((PangoAttribute *)iterator->next_attribute->data)->end_index > iterator->start_index)
+	{
+	  iterator->attribute_stack = g_list_prepend (iterator->attribute_stack, iterator->next_attribute->data);
+	  iterator->end_index = MIN (iterator->end_index, ((PangoAttribute *)iterator->next_attribute->data)->end_index);
+	}
       iterator->next_attribute = iterator->next_attribute->next;
     }
 
@@ -1370,7 +1429,7 @@ pango_attr_iterator_get (PangoAttrIterator *iterator,
  *        so if you plan to keep it around, you must call:
  *        <literal>pango_font_description_set_family (desc, pango_font_description_get_family (desc))</literal>.
  * @language: if non-%NULL, location to store language tag for item, or %NULL
- *            if non is found.
+ *            if none is found.
  * @extra_attrs: if non-%NULL, location in which to store a list of non-font
  *           attributes at the the current position; only the highest priority
  *           value of each attribute will be added to this list. In order
